using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Windows.Forms;

namespace Proyecto0
{
    /// <sumary>
    ///   Clase responsable de jugar para la CPU
    /// </sumary>
    internal class CPU
    {
        // Tablero temporal usado para los movimientos de la CPU
        private Tablero tableroTemp;

        // Color de las fichas de la CPU
        //private int color;

        // Profundidad mxima usada en el MiniMax (por defecto, 2)
        private int profundidadMax = 2;

        // Pesos usados para el tablero
        private int[] pesosTablero = { 4, 4, 4, 4, 
                                 4, 3, 3, 3,
                                 3, 2, 2, 4,
                                 4, 2, 1, 3,
                                 3, 1, 2, 4,
                                 4, 2, 2, 3,
                                 3, 3, 3, 4,
                                 4, 4, 4, 4};




        /// <sumary>
        ///  Constructor principal de la clase
        /// </sumary>
        /// <param name="t">
        ///  Tablero que la CPU utiliza para calcular las jugadas
        /// </param>
        public CPU(Tablero t)
        {
            tableroTemp = t;
            //color = Tablero.NEGRA;
        }

        /// <sumary> 
        ///   Permite al usuario cambiar la profundidad
        /// </sumary>
        public int profundidad
        {
            get
            {
                return profundidadMax;
            }
            set
            {
                profundidadMax = value;
            }
        }

        /// <sumary> 
        ///   Ordena a la CPU a realizar un movimiento
        /// </sumary>
        public void play()
        {
            try
            {
                Lista movimientos = minimax(tableroTemp);
                Enumeration iter = movimientos.elements();
                Movimiento movimiento;
                int filaOrigen;
                int columnaOrigen;
                int filaDestino;
                int columnaDestino;

                if (!movimientos.estaVacia())
                {
                    // 14/10/2008
                    // CAMBIO PARA INDICAR LOS MOVIMIENTOS DE LA CPU
                    Debug.WriteLine("Movimientos de CPU:\n");
                    while (iter.hasMoreElements())
                    {
                        movimiento = (Movimiento)iter.nextElement();
                        filaOrigen = (movimiento.getFrom() / 4) + 1;
                        columnaOrigen = ((movimiento.getFrom() % 4) * 2 + ((movimiento.getFrom() / 4) % 2 == 0 ? 1 : 0)) + 1;
                        filaDestino = (movimiento.getTo() / 4) + 1;
                        columnaDestino = ((movimiento.getTo() % 4) * 2 + ((movimiento.getTo() / 4) % 2 == 0 ? 1 : 0)) + 1;
                        Debug.WriteLine("Salto de (" + filaOrigen + "," + columnaOrigen + ") a (" + filaDestino + "," + columnaDestino + ")\n");
                    }
                    tableroTemp.mover(movimientos);
                }
            }
            catch (MovInvalido bad)
            {
                Debug.WriteLine(bad.StackTrace);
                Application.Exit();
            }
        }

        /// <sumary> 
        ///   Cambia el tablero temporal
        /// </sumary>
        /// <param name="t">
        ///  Nuevo tablero
        /// </param>
        public void setTablero(Tablero t)
        {
            tableroTemp = t;
        }

        /// <sumary> 
        ///   Comprueba si la lista de movimientos es vlida
        /// </sumary>
        /// <param name="moves">
        ///  Lista de movimientos
        /// </param>
        /// <value>
        ///  Devuelve cierto si el movimiento es vlido y falso en caso contrario
        /// </value>
        private bool puedeMover(Lista moves)
        {
            return !moves.estaVacia() && !((Lista)moves.peekCabeza()).estaVacia();
        }


        /// <sumary> 
        ///   Mtodo que implementa el algoritmo MiniMax para decidir el movimiento
        /// </sumary>
        /// <param name="t">
        ///   Tablero sobre el que se generar el movimiento 
        /// </param>
        /// <value>
        ///  Lista con los movimientos que ha calculado el Minimax para la CPU
        /// </value>
        private Lista minimax(Tablero t)
        {
            Lista sucesores;
            Lista movimiento, mejorMov = null;
            Tablero siguienteTablero;
            int valor, maxValor = Int32.MinValue;

            sucesores = t.movimientosValidos();
            while (puedeMover(sucesores))
            {
                movimiento = (Lista)sucesores.extraerCabeza();
                siguienteTablero = (Tablero)t.clone();

                Debug.WriteLine("******************************************************************");
                siguienteTablero.mover(movimiento);
                valor = movimientoMAX(siguienteTablero, 1, maxValor, Int32.MaxValue);

                if (valor > maxValor)
                {
                    Debug.WriteLine("Valor mximo: " + valor + " a profundidad : 0");
                    maxValor = valor;
                    mejorMov = movimiento;
                }
            }

            Debug.WriteLine("Valor seleccionado: " + maxValor + " a profundidad : 0");

            return mejorMov;
        }

        /// <sumary> 
        ///   Implementa la eleccin de los nodos MAX
        /// </sumary>
        /// <param name="tablero">
        ///   Tablero del que se parte para generar los movimientos
        /// </param>
        /// <param name="profundidad">
        ///   Profundidad que lleva el rbol minimax
        /// </param>
        /// <param name="alpha">
        ///   Valor alpha de la poda alpha-beta
        /// </param>
        /// <param name="beta">
        ///   Valor beta de la poda alpha-beta
        /// </param>
        /// <value>
        ///  Valor elegido
        /// </value>
        private int movimientoMAX(Tablero tablero, int profundidad, int alpha, int beta)
        {
            if (aplicarPoda(tablero, profundidad))
                return eval(tablero);


            Lista sucesores;
            Lista movimiento;
            Tablero siguienteTablero;
            int valor;

            Debug.WriteLine("Nodo MAX a profundidad : " + profundidad + " con alpha : " + alpha +
                                " beta : " + beta);

            sucesores = tablero.movimientosValidos();
            while (puedeMover(sucesores))
            {
                movimiento = (Lista)sucesores.extraerCabeza();
                siguienteTablero = (Tablero)tablero.clone();
                siguienteTablero.mover(movimiento);
                valor = movimientoMIN(siguienteTablero, profundidad + 1, alpha, beta);

                if (valor > alpha)
                {
                    alpha = valor;
                    Debug.WriteLine("Nodo MAX : " + valor + " a profundidad : " + profundidad);
                }

                if (alpha > beta)
                {
                    Debug.WriteLine("Nodo MAX con poda : " + beta + " a profundidad : " + profundidad);
                    Debug.WriteLine(sucesores.longitud() + " sucesores cortados");
                    return beta;
                }

            }

            Debug.WriteLine("Nodo MAX seleccionado : " + alpha + " a profundidad : " + profundidad);
            return alpha;
        }

        /// <sumary> 
        ///   Implementa la eleccin de los nodos MIN
        /// </sumary>
        /// <param name="tablero">
        ///   Tablero del que se parte para generar los movimientos
        /// </param>
        /// <param name="depth">
        ///   Profundidad que lleva el rbol minimax
        /// </param>
        /// <param name="alpha">
        ///   Valor alpha de la poda alpha-beta
        /// </param>
        /// <param name="beta">
        ///   Valor beta de la poda alpha-beta
        /// </param>
        /// <value>
        ///  Valor elegido
        /// </value>
        private int movimientoMIN(Tablero tablero, int profundidad, int alpha, int beta)
        {
            if (aplicarPoda(tablero, profundidad))
                return eval(tablero);


            Lista sucesores;
            Lista movimiento;
            Tablero siguienteTablero;
            int valor;

            Debug.WriteLine("Nodo MIN a profundidad : " + profundidad + " con alpha : " + alpha +
                                " beta : " + beta);

            sucesores = (Lista)tablero.movimientosValidos();
            while (puedeMover(sucesores))
            {
                movimiento = (Lista)sucesores.extraerCabeza();
                siguienteTablero = (Tablero)tablero.clone();
                siguienteTablero.mover(movimiento);
                valor = movimientoMAX(siguienteTablero, profundidad + 1, alpha, beta);

                if (valor < beta)
                {
                    beta = valor;
                    Debug.WriteLine("Valor MIN : " + valor + " a profundidad : " + profundidad);
                }

                if (beta < alpha)
                {
                    Debug.WriteLine("Valor MIN con poda : " + alpha + " a profundidad : " + profundidad);
                    Debug.WriteLine(sucesores.longitud() + " sucesores cortados");
                    return alpha;
                }
            }

            Debug.WriteLine("Valor MIN seleccionado : " + beta + " a profundidad : " + profundidad);
            return beta;
        }

        /// <sumary> 
        ///   Funcin de evaluacin esttica para las damas
        /// </sumary>
        /// <param name="tablero">
        ///   El tablero sobre el que se evaluara la posicin del jugador
        /// </param>
        /// <value>
        ///  Valor obtenido
        /// </value>
        private int eval(Tablero tablero)
        {
            
            int peligrosidadCPU = 0;
            int peligrosidadJugador = 0;
            int ficha;

            try
            {
                for (int i = 0; i < 32; i++)
                {
                    ficha = tablero.getFicha(i);

                    if (ficha != Tablero.VACIA)
                    {
                        if (ficha == Tablero.NEGRA)
                            peligrosidadCPU += 1;
                        else if (ficha == Tablero.REINA_NEGRA)
                            peligrosidadCPU += 3;
                        else if (ficha == Tablero.BLANCA)
                            peligrosidadJugador += 1;
                        else if (ficha == Tablero.REINA_BLANCA)
                            peligrosidadJugador += 3;
                    }
                }
            }
            catch (CoordErronea e)
            {
                Debug.WriteLine(e.StackTrace);
                Application.Exit();
            }

            return peligrosidadCPU - peligrosidadJugador;

            /*
            try
            {
                for (int i = 0; i < 32; i++)
                {
                    ficha = tablero.getFicha(i);

                    if (ficha != Tablero.VACIA)
                        if (ficha == color || ficha == colorReina)
                            peligrosidadCPU += calcularValor(ficha, i);
                        else
                            peligrosidadJugador += calcularValor(ficha, i);
                }
            }
            catch (CoordErronea e)
            {
                Debug.WriteLine(bad.StackTrace);
                Application.Exit();
            }

            return peligrosidadCPU - peligrosidadJugador;
             */
            }

        /// <sumary> 
        ///   Funcin de evaluacin para las fichas
        /// </sumary>
        /// <param name="ficha">
        ///   Tipo de ficha
        /// </param>
        /// <param name="pos">
        ///   Posicin de la ficha
        /// </param>
        /// <value>
        ///  Valor de evaluacin
        /// </value>
        private int calcularValor(int ficha, int pos)
        {
            int valor;

            if (ficha == Tablero.BLANCA) // Ficha simple blanca
                if (pos >= 4 && pos <= 7)
                    valor = 7;
                else
                    valor = 5;
            else if (ficha != Tablero.NEGRA) // Ficha simple negra
                if (pos >= 24 && pos <= 27)
                    valor = 7;
                else
                    valor = 5;
            else // Ficha reina
                valor = 10;

            return valor * pesosTablero[pos];
        }


        /// <sumary> 
        ///   Comprueba si el rbol debe ser podado
        /// </sumary>
        /// <param name="tablero">
        ///   Tablero que se est procesando
        /// </param>
        /// <param name="profundidad">
        ///   Profundidad actual del rbol
        /// </param>
        /// <value>
        ///  Devuelve true si el rbol puede ser podado
        /// </value>
        private bool aplicarPoda(Tablero tablero, int profundidad)
        {
            return profundidad > profundidadMax || tablero.juegoTerminado();
        }

    }
}
