namespace Proyecto0
{

    using System;

    /// <sumary>
    ///   Clase que representa el tablero de juego
    /// </sumary>    

    public class Tablero
    {
        // Representacin del tablero mediante un array de byte
        private byte[] tab;

        // Constantes para las fichas
        public const byte VACIA = 0;
        public const byte BLANCA = 2;
        public const byte REINA_BLANCA = 3;
        public const byte NEGRA = 4;
        public const byte REINA_NEGRA = 5;

        // Constante utiliza para comprobar si una ficha es reina
        private const byte REINA = 1;

        // Nmero de fichas blancas que quedan
        private int fichasBlancas;
        // Nmero de fichas negras que quedan
        private int fichasNegras;


        // Jugador que tiene el turno
        private int jugadorConTurno;


        /// <sumary>
        ///   Constructor principal de la clase
        /// </sumary>    
        public Tablero()
        {
            tab = new byte[32];
            inicializarTablero();
        }


        /// <sumary>
        ///   Devuelve el jugador que tiene el turno
        /// </sumary>    
        public int getJugadorConTurno()
        {
            return jugadorConTurno;
        }

        /// <sumary>
        ///   Cambia el jugador que lleva el turno
        /// </sumary>    
        
        public void setJugadorConTurno(int jugador)
        {
            jugadorConTurno = jugador;
        }

        /// <sumary>
        ///   Devuelve el nmero de fichas blancas
        /// </sumary>
        public int getFichasBlancas()
        {
            return fichasBlancas;
        }

        /// <sumary>
        ///   Devuelve el nmero de fichas negras
        /// </sumary>   
        public int getFichasNegras()
        {
            return fichasNegras;
        }

        /// <sumary>
        ///   Devuelve un clon del objeto Tablero instanciado
        /// </sumary> 
        public object clone()
        {
            Tablero tablero = new Tablero();

            tablero.jugadorConTurno = jugadorConTurno;
            tablero.fichasBlancas = fichasBlancas;
            tablero.fichasNegras = fichasNegras;
            for (int i = 0; i < 32; i++)
                tablero.tab[i] = tab[i];

            return tablero;
        }



        /// <sumary>
        ///   Devuelve una lista con todos los movimientos vlidos posibles para
        ///   el jugador con el turno sobre el tablero actual
        /// </sumary>    
        public Lista movimientosValidos()
        {
            int color;
            int jugadorContrario;

            color = jugadorConTurno;
            if (color == BLANCA)
                jugadorContrario = NEGRA;
            else
                jugadorContrario = BLANCA;

            if (debeAtacar())
                return generaMovimientosAtaque(color, jugadorContrario);
            else
                return generaMovimientos(color, jugadorContrario);
        }

        /// <sumary>
        ///   Devuelve una lista con todos los movimientos de ataque vlidos
        ///   posibles para el jugador con el color indicado sobre el tablero actual
        /// </sumary>    
        /// <param name="colorTurno">
        ///  Color de las fichas del jugador con el turno 
        /// </param>
        /// <param name="jugadorContrario">
        ///  Color de las fichas del jugador contrario
        /// </param>
        /// <value>
        ///  Devuelve una lista con todos los movimientos de ataque vlidos
        /// </value>
        private Lista generaMovimientosAtaque(int colorTurno, int jugadorContrario)
        {
            Lista movimientos = new Lista();
            Lista movimientosTemp;


            for (int k = 0; k < 32; k++)
                if ((tab[k] & ~REINA) == jugadorConTurno)
                {
                    if ((tab[k] & REINA) == 0)  // Ficha simple
                        movimientosTemp = ataquesSimples(k, colorTurno, jugadorContrario);
                    else
                    { // Ficha reina
                        Lista ultimaPosicion = new Lista();

                        ultimaPosicion.insertarCola(k);

                        movimientosTemp = ataquesReina(ultimaPosicion, k, NADA, colorTurno, jugadorContrario);
                    }

                    if (noVacia(movimientosTemp))
                        movimientos.concatenar(movimientosTemp);
                }

            return movimientos;
        }




        /// <sumary>
        ///   Devuelve una lista con todos los movimientos vlidos posibles
        ///   utilizando fichas simples desde una posicin indicada
        /// </sumary>   
        /// <param name="pos">
        ///  Posicin de la ficha actual
        /// </param>    
        /// <param name="colorTurno">
        ///  Color del jugador con turno
        /// </param>
        /// <param name="jugadorContrario">
        ///  Color del del jugador contrario
        /// </param>
        /// <value>
        ///  Lista de movimientos
        /// </value>
        private Lista ataquesSimples(int pos, int colorTurno, int jugadorContrario)
        {
            int x = posToCol(pos);
            int y = posToFila(pos);
            int i;
            Lista movimientos = new Lista();
            Lista movimientosTemp;
            int posContrario, posSiguiente;



            i = (colorTurno == BLANCA) ? -1 : 1;


            // Comprueba las diagonales / hacia arriba y \ hacia abajo
            if (x < 6 && y + i > 0 && y + i < 7)
            {
                posContrario = colLineToPos(x + 1, y + i);
                posSiguiente = colLineToPos(x + 2, y + 2 * i);

                if ((tab[posContrario] & ~REINA) == jugadorContrario && tab[posSiguiente] == VACIA)
                {
                    movimientosTemp = ataquesSimples(posSiguiente, colorTurno, jugadorContrario);
                    movimientos.concatenar(anadirMovimiento(new Movimiento(pos, posSiguiente), movimientosTemp));
                }
            }


            // Comprueba las diagonales / hacia abajo y \ hacia arriba
            if (x > 1 && y + i > 0 && y + i < 7)
            {
                posContrario = colLineToPos(x - 1, y + i);
                posSiguiente = colLineToPos(x - 2, y + 2 * i);

                if ((tab[posContrario] & ~REINA) == jugadorContrario && tab[posSiguiente] == VACIA)
                {
                    movimientosTemp = ataquesSimples(posSiguiente, colorTurno, jugadorContrario);
                    movimientos.concatenar(anadirMovimiento(new Movimiento(pos, posSiguiente), movimientosTemp));
                }
            }

            if (movimientos.estaVacia())
                movimientos.insertarCola(new Lista());

            return movimientos;
        }


        // Constantes que indican la direccin del ltimo movimiento hecho por la reina
        // durante la bsqueda de los movimientos (para no repetir la direccin)
        private const int NADA = 0;        // Sin mover
        private const int IZQUIERDA_ABAJO = 1; // Diagonal / hacia abajo
        private const int IZQUIERDA_ARRIBA = 2; // Diagonal \ hacia arriba
        private const int DERECHA_ABAJO = 3; // Diagonal \ hacia abajo
        private const int DERECHA_ARRIBA = 4; // Diagonal / hacia arriba


        /// <sumary>
        ///   Devuelve una lista con todos los movimientos vlidos posibles
        ///   utilizando ficha reina desde una posicin indicada
        /// </sumary>   
        /// <param name="movReina">
        ///   Lista que contiene todos los movimientos virtuales hechos por la reina hasta ahora
        /// </param>
        /// <param name="pos">
        ///   Posicin actual
        /// </param>
        /// <param name="dir">
        ///   ltima direccin
        /// </param>    
        /// <param name="colorTurno">
        ///  Color del jugador con turno
        /// </param>
        /// <param name="jugadorContrario">
        ///  Color del jugador contrario
        /// </param>
        /// <value>
        ///  Lista de movimientos
        /// </value>
        private Lista ataquesReina(Lista movReina, int pos, int dir, int colorTurno, int jugadorContrario)
        {
            Lista movimientosTemp, movimientos = new Lista();

            if (dir != DERECHA_ABAJO)
            {
                movimientosTemp = ataquesDiagonalReina(movReina, pos, colorTurno, jugadorContrario, 1, 1);

                if (noVacia(movimientosTemp))
                    movimientos.concatenar(movimientosTemp);
            }

            if (dir != IZQUIERDA_ARRIBA)
            {
                movimientosTemp = ataquesDiagonalReina(movReina, pos, colorTurno, jugadorContrario, -1, -1);

                if (noVacia(movimientosTemp))
                    movimientos.concatenar(movimientosTemp);
            }


            if (dir != DERECHA_ARRIBA)
            {
                movimientosTemp = ataquesDiagonalReina(movReina, pos, colorTurno, jugadorContrario, 1, -1);

                if (noVacia(movimientosTemp))
                    movimientos.concatenar(movimientosTemp);
            }

            if (dir != IZQUIERDA_ABAJO)
            {
                movimientosTemp = ataquesDiagonalReina(movReina, pos, colorTurno, jugadorContrario, -1, 1);

                if (noVacia(movimientosTemp))
                    movimientos.concatenar(movimientosTemp);
            }


            return movimientos;
        }


        /// <sumary>
        ///   Devuelve una lista con todos los movimientos vlidos posibles
        ///   utilizando ficha reina para la diagonal indicada
        /// </sumary>
        /// <param name="movReina">
        ///   Lista que contiene todos los movimientos virtuales hechos por la reina hasta ahora
        /// </param>
        /// <param name="pos">
        ///   Posicin actual
        /// </param>
        /// <param name="colorTurno">
        ///  Color del jugador con turno
        /// </param>
        /// <param name="jugadorContrario">
        ///  Color del jugador contrario
        /// </param>
        /// <param name="dirX">
        ///  Color del jugador contrario
        /// </param>
        /// <param name="dirY">
        ///  Color del jugador contrario
        /// </param>
        /// <value>
        ///  Lista de movimientos
        /// </value>
        private Lista ataquesDiagonalReina(Lista movReina, int pos, int colorTurno, int jugadorContrario, int dirX, int dirY)
        {
            int x = posToCol(pos);
            int y = posToFila(pos);
            int i, j;
            Lista movimientos = new Lista();
            Lista movimientosTemp, tempPos;


            int posInicial = (int)movReina.peekCabeza();

            i = x + dirX;
            j = y + dirY;

            // Busca el jugador contrario
            while (i > 0 && i < 7 && j > 0 && j < 7 &&
                   (tab[colLineToPos(i, j)] == VACIA || colLineToPos(i, j) == posInicial))
            {
                i += dirX;
                j += dirY;
            }

            if (i > 0 && i < 7 && j > 0 && j < 7 && (tab[colLineToPos(i, j)] & ~REINA) == jugadorContrario &&
                !movReina.contiene(colLineToPos(i, j)))
            {

                movReina.insertarCola(colLineToPos(i, j));

                i += dirX;
                j += dirY;

                int i_aux = i;
                int j_aux = j;
                while (i >= 0 && i <= 7 && j >= 0 && j <= 7 &&
                     (tab[colLineToPos(i, j)] == VACIA || colLineToPos(i, j) == posInicial))
                {

                    int dir;

                    if (dirX == 1 && dirY == 1)
                        dir = IZQUIERDA_ARRIBA;
                    else if (dirX == -1 && dirY == -1)
                        dir = DERECHA_ABAJO;
                    else if (dirX == -1 && dirY == 1)
                        dir = DERECHA_ARRIBA;
                    else
                        dir = IZQUIERDA_ABAJO;


                    tempPos = (Lista)movReina.clone();
                    movimientosTemp = ataquesReina(tempPos, colLineToPos(i, j), dir, colorTurno, jugadorContrario);

                    if (noVacia(movimientosTemp))
                        movimientos.concatenar(anadirMovimiento(new Movimiento(pos, colLineToPos(i, j)), movimientosTemp));

                    i += dirX;
                    j += dirY;
                }

                movReina.extraerCola();

                if (movimientos.estaVacia())
                {
                    i = i_aux;
                    j = j_aux;

                    while (i >= 0 && i <= 7 && j >= 0 && j <= 7 &&
                           (tab[colLineToPos(i, j)] == VACIA || colLineToPos(i, j) == posInicial))
                    {

                        movimientosTemp = new Lista();
                        movimientosTemp.insertarCola(new Movimiento(pos, colLineToPos(i, j)));
                        movimientos.insertarCola(movimientosTemp);

                        i += dirX;
                        j += dirY;
                    }
                }
            }

            return movimientos;
        }


        /// <sumary>
        ///   Devuelve true si la lista de movimientos pasada como parmetro no est vaca
        ///   y contiene un movimiento
        /// </sumary>
        private bool noVacia(Lista movimientos)
        {
            return !movimientos.estaVacia() && !((Lista)movimientos.peekCabeza()).estaVacia();
        }

        /// <sumary>
        ///   Aade un nuevo movimiento a la cabeza de la lista de movimientos y la devuelve
        /// </sumary>
        /// <param name="movimiento">
        ///   Movimiento a aadir a la lista
        /// </param>
        /// <param name="movimientos">
        ///   Lista de movimientos
        /// </param>
        /// <value>
        ///  Lista de movimientos actualizada con nuevo movimiento
        /// </value>
        private Lista anadirMovimiento(Movimiento movimiento, Lista movimientos)
        {
            if (movimiento == null)
                return movimientos;

            Lista movActual, listaRes = new Lista();
            while (!movimientos.estaVacia())
            {
                movActual = (Lista)movimientos.extraerCabeza();
                movActual.insertarCabeza(movimiento);
                listaRes.insertarCola(movActual);
            }

            return listaRes;
        }



        /// <sumary>
        ///   Devuelve una lista con todos los movimientos vlidos posibles que no sean de ataque
        ///   para el jugador con el color indicado sobre el tablero actual
        /// </sumary>    
        /// <param name="colorTurno">
        ///  Color de las fichas del jugador con el turno 
        /// </param>
        /// <param name="jugadorContrario">
        ///  Color de las fichas del jugador contrario
        /// </param>
        /// <value>
        ///  Devuelve una lista con todos los movimientos de ataque vlidos
        /// </value>
        private Lista generaMovimientos(int colorTurno, int jugadorContrario)
        {
            Lista movimientos = new Lista();
            Lista movimientosTemp;


            for (int k = 0; k < 32; k++)
                if ((tab[k] & ~REINA) == jugadorConTurno)
                {
                    int x = posToCol(k);
                    int y = posToFila(k);
                    int i, j;

                    if ((tab[k] & REINA) == 0)
                    {  // Ficha simple
                        i = (colorTurno == BLANCA) ? -1 : 1;

                        // Comprueba las diagonales / hacia arriba y \ hacia abajo
                        if (x < 7 && y + i >= 0 && y + i <= 7 &&
                            tab[colLineToPos(x + 1, y + i)] == VACIA)
                        {
                            movimientosTemp = new Lista();
                            movimientosTemp.insertarCola(new Movimiento(k, colLineToPos(x + 1, y + i)));
                            movimientos.insertarCola(movimientosTemp);
                        }


                        // Comprueba las diagonales \ hacia arriba y / hacia abajo
                        if (x > 0 && y + i >= 0 && y + i <= 7 &&
                            tab[colLineToPos(x - 1, y + i)] == VACIA)
                        {
                            movimientosTemp = new Lista();
                            movimientosTemp.insertarCola(new Movimiento(k, colLineToPos(x - 1, y + i)));
                            movimientos.insertarCola(movimientosTemp);
                        };
                    }
                    else
                    { // Ficha reina
                        // Comprueba la diagonal \ hacia abajo
                        i = x + 1;
                        j = y + 1;

                        while (i <= 7 && j <= 7 && tab[colLineToPos(i, j)] == VACIA)
                        {
                            movimientosTemp = new Lista();
                            movimientosTemp.insertarCola(new Movimiento(k, colLineToPos(i, j)));
                            movimientos.insertarCola(movimientosTemp);

                            i++;
                            j++;
                        }


                        // Comprueba la diagonal \ hacia arriba
                        i = x - 1;
                        j = y - 1;
                        while (i >= 0 && j >= 0 && tab[colLineToPos(i, j)] == VACIA)
                        {
                            movimientosTemp = new Lista();
                            movimientosTemp.insertarCola(new Movimiento(k, colLineToPos(i, j)));
                            movimientos.insertarCola(movimientosTemp);

                            i--;
                            j--;
                        }

                        // Comprueba la diagonal / hacia arriba
                        i = x + 1;
                        j = y - 1;
                        while (i <= 7 && j >= 0 && tab[colLineToPos(i, j)] == VACIA)
                        {
                            movimientosTemp = new Lista();
                            movimientosTemp.insertarCola(new Movimiento(k, colLineToPos(i, j)));
                            movimientos.insertarCola(movimientosTemp);

                            i++;
                            j--;
                        }

                        // Comprueba la diagonal \ hacia abajo
                        i = x - 1;
                        j = y + 1;
                        while (i >= 0 && j <= 7 && tab[colLineToPos(i, j)] == VACIA)
                        {
                            movimientosTemp = new Lista();
                            movimientosTemp.insertarCola(new Movimiento(k, colLineToPos(i, j)));
                            movimientos.insertarCola(movimientosTemp);

                            i--;
                            j++;
                        }
                    }
                }

            return movimientos;
        }

        /// <sumary>
        ///   Valida un movimiento
        /// </sumary>
        /// <param name="from">
        ///   Posicin inicial
        /// </param>
        /// <param name="to">
        ///   Posicin final
        /// </param>
        /// <value>
        ///  Devuelve true si el movimiento es vlido
        /// </value>
        public bool esMovimientoValido(int from, int to)
        {
            // Argumentos errneos
            if (from < 0 || from > 32 || to < 0 || to > 32)
                return false;

            // Posicin inicial vaca o posicin final ocupada
            if (tab[from] == VACIA || tab[to] != VACIA)
            {   
                // Movimiento invlido de casilla ocupada
                Partida.tipoMovInvalido = 2;
                return false;
            }
                

            // Intentando mover ficha de jugador contrario
            if ((tab[from] & ~REINA) != jugadorConTurno)
                return false;


            int colorTurno;
            int jugadorContrario;
            colorTurno = tab[from] & ~REINA;
            if (colorTurno == BLANCA)
                jugadorContrario = NEGRA;
            else
                jugadorContrario = BLANCA;


            int fromLinea = posToFila(from);
            int fromCol = posToCol(from);
            int toLinea = posToFila(to);
            int toCol = posToCol(to);

            int incX, incY;

            // Inicializamos incrementos
            if (fromCol > toCol)
                incX = -1;
            else
                incX = 1;


            if (fromLinea > toLinea)
                incY = -1;
            else
                incY = 1;

            int x = fromCol + incX;
            int y = fromLinea + incY;


            if ((tab[from] & REINA) == 0)
            { // Ficha simple
                bool buenaDireccion;

                if ((incY == -1 && colorTurno == BLANCA) || (incY == 1 && colorTurno == NEGRA))
                    buenaDireccion = true;
                else
                {
                    // Movimiento invlido de no retroceso
                    Partida.tipoMovInvalido = 3;
                    buenaDireccion = false;
                    return buenaDireccion;
                }

                if (x == toCol && y == toLinea)
                {
                    // Movimiento simple
                    Boolean invalido = debeAtacar();
                    if (invalido)
                        // Movimiento invlido: Debes comer ficha
                        Partida.tipoMovInvalido = 4;
                    return buenaDireccion && !invalido;
                }



                // Si no es un movimiento simple debe ser un movimiento de ataque
                return buenaDireccion && x + incX == toCol && y + incY == toLinea &&
                       (tab[colLineToPos(x, y)] & ~REINA) == jugadorContrario;
            }
            else
            { // Ficha reina
                while (x != toCol && y != toLinea && tab[colLineToPos(x, y)] == VACIA)
                {
                    x += incX;
                    y += incY;
                }

                // Movimiento simple con la reina
                if (x == toCol && y == toLinea)
                    return !debeAtacar();

                // Movimiento de ataque con la reina
                if ((tab[colLineToPos(x, y)] & ~REINA) == jugadorContrario)
                {
                    x += incX;
                    y += incY;

                    // Para que las reinas puedan volar
                    while (x != toCol && y != toLinea && tab[colLineToPos(x, y)] == VACIA)
                    {
                        x += incX;
                        y += incY;
                    }

                    if (x == toCol && y == toLinea)
                        return true;
                }
            }


            return false;
        }


        /// <sumary>
        ///   Comprueba si el jugador en turno debe atacar
        /// </sumary>
        /// <value>
        ///  Devuelve true si hay fichas enemigas que debe atacar
        /// </value>
        public bool debeAtacar()
        {
            for (int i = 0; i < 32; i++)
                if ((tab[i] & ~REINA) == jugadorConTurno && puedeAtacar(i))
                    return true;

            return false;
        }

        /// <sumary>
        ///   Comprueba si se puede atacar desde una posicin pasada como parmetro
        /// </sumary>
        /// <param name="pos">
        ///   Posicin de tablero
        /// </param>
        /// <value>
        ///  Devuelve true si hay una ficha en esa posicin y puede atacar a otra ficha
        /// </value>
        public bool puedeAtacar(int pos)
        {
            if (tab[pos] == VACIA)
                return false;

            int colorTurno;
            int jugadorContrario;

            colorTurno = tab[pos] & ~REINA;
            if (colorTurno == BLANCA)
                jugadorContrario = NEGRA;
            else
                jugadorContrario = BLANCA;

            int x = posToCol(pos);
            int y = posToFila(pos);

            if ((tab[pos] & REINA) == 0)
            { // Ficha simple
                int i;

                i = (colorTurno == BLANCA) ? -1 : 1;

                // Diagonales / hacia arriba y \ hacia abajo
                if (x < 6 && y + i > 0 && y + i < 7 && (tab[colLineToPos(x + 1, y + i)] & ~REINA) == jugadorContrario &&
                    tab[colLineToPos(x + 2, y + 2 * i)] == VACIA)
                    return true;

                // Diagonales \ hacia arriba y / hacia abajo
                if (x > 1 && y + i > 0 && y + i < 7 && (tab[colLineToPos(x - 1, y + i)] & ~REINA) == jugadorContrario &&
                    tab[colLineToPos(x - 2, y + 2 * i)] == VACIA)
                    return true;

            }
            else
            { // Ficha reina
                int i, j;


                // Diagonal \ hacia abajo
                i = x + 1;
                j = y + 1;
                while (i < 6 && j < 6 && tab[colLineToPos(i, j)] == VACIA)
                {
                    i++;
                    j++;
                }

                if (i < 7 && j < 7 && (tab[colLineToPos(i, j)] & ~REINA) == jugadorContrario)
                {
                    i++;
                    j++;

                    if (i <= 7 && j <= 7 && tab[colLineToPos(i, j)] == VACIA)
                        return true;
                }

                // Diagonal \ hacia arriba
                i = x - 1;
                j = y - 1;
                while (i > 1 && j > 1 && tab[colLineToPos(i, j)] == VACIA)
                {
                    i--;
                    j--;
                }

                if (i > 0 && j > 0 && (tab[colLineToPos(i, j)] & ~REINA) == jugadorContrario)
                {
                    i--;
                    j--;

                    if (i >= 0 && j >= 0 && tab[colLineToPos(i, j)] == VACIA)
                        return true;
                }

                // Diagonal / hacia arriba
                i = x + 1;
                j = y - 1;
                while (i < 6 && j > 1 && tab[colLineToPos(i, j)] == VACIA)
                {
                    i++;
                    j--;
                }

                if (i < 7 && j > 0 && (tab[colLineToPos(i, j)] & ~REINA) == jugadorContrario)
                {
                    i++;
                    j--;

                    if (i <= 7 && j >= 0 && tab[colLineToPos(i, j)] == VACIA)
                        return true;
                }

                // Diagonal / hacia abajo
                i = x - 1;
                j = y + 1;
                while (i > 1 && j < 6 && tab[colLineToPos(i, j)] == VACIA)
                {
                    i--;
                    j++;
                }

                if (i > 0 && j < 7 && (tab[colLineToPos(i, j)] & ~REINA) == jugadorContrario)
                {
                    i--;
                    j++;

                    if (i >= 0 && j <= 7 && tab[colLineToPos(i, j)] == VACIA)
                        return true;
                }
            }


            return false;
        }

        /// <sumary>
        ///   Ejecuta un nico movimiento desde la posicin inicial a la posicin final
        /// </sumary>
        /// <param name="from">
        ///   Posicin inicial
        /// </param>
        /// <param name="to">
        ///   Posicin final
        /// </param>
        public void mover(int from, int to)
        {
            bool movMultiple = debeAtacar();

            aplicarMovimiento(from, to);

            if (!movMultiple)
                cambiarTurno();
            else
                if (!puedeAtacar(to))
                    cambiarTurno();
        }

        /// <sumary>
        ///   Ejecuta un movimiento mltiple (lista de movimientos)
        /// </sumary>
        /// <param name="movimientos">
        ///   Lista con todos los movimientos a ejecutar
        /// </param>
        public void mover(Lista movimientos)
        {
            Movimiento movimiento;
            Enumeration iter = movimientos.elements();

            while (iter.hasMoreElements())
            {
                movimiento = (Movimiento)iter.nextElement();
                aplicarMovimiento(movimiento.getFrom(), movimiento.getTo());                
            }

            cambiarTurno();
        }


        /// <sumary>
        ///   Cambia el turno de jugador
        /// </sumary>
        private void cambiarTurno()
        {
            if (jugadorConTurno == BLANCA)
                jugadorConTurno = NEGRA;
            else
                jugadorConTurno = BLANCA;
        }


        /// <sumary>
        ///   Ejecuta un movimiento y cambia la informacin del tablero
        /// </sumary>
        /// <param name="from">
        ///   Posicin inicial
        /// </param>
        /// <param name="to">
        ///   Posicin final
        /// </param>
        private void aplicarMovimiento(int from, int to) 
        {   
            if (!esMovimientoValido(from, to))
                throw new MovInvalido();

            limpiarFichas(from, to);
            // Ejecutamos el movimiento
            if (to < 4 && tab[from] == BLANCA)
                tab[to] = REINA_BLANCA;
            else if (to > 27 && tab[from] == NEGRA)
                tab[to] = REINA_NEGRA;
            else
                tab[to] = tab[from];

            tab[from] = VACIA;
        }


        /// <sumary>
        ///   Devuelve la ficha en la posicin indicada
        /// </sumary>
        /// <param name="pos">
        ///   La posicin de la ficha en el tablero
        /// </param>
        /// <value>
        ///   Ficha en la posicin indicada
        /// </value>
        public byte getFicha(int pos)
        {
            if (pos < 0 || pos > 32)
                throw new CoordErronea();

            return tab[pos];
        }

        /// <sumary>
        ///   Comprueba si la partida ha terminado
        /// </sumary>
        /// <value>
        ///   Devuelve true si la partida ha terminado
        /// </value>
        public bool juegoTerminado()
        {
            return fichasBlancas == 0 || fichasNegras == 0 || !noVacia(movimientosValidos());
        }


        /// <sumary>
        ///   Nos dice que jugador ha ganado
        /// </sumary>
        /// <value>
        ///   Devuelve un entero que indica el color ganador
        /// </param>
        public int ganador()
        {
            if (jugadorConTurno == BLANCA)
                if (noVacia(movimientosValidos()))
                    return BLANCA;
                else
                    return NEGRA;
            else if (noVacia(movimientosValidos()))
                return NEGRA;
            else
                return BLANCA;
        }



        /// <sumary>
        ///   Elimina las fichas en el tablero entre una posicin inicial y final
        /// </sumary>
        /// <param name="from">
        ///   Posicin final
        /// </param>
        /// <param name="to">
        ///   Posicin final
        /// </param>
        private void limpiarFichas(int from, int to)
        {
            int fromLinea = posToFila(from);
            int fromCol = posToCol(from);
            int toLinea = posToFila(to);
            int toCol = posToCol(to);

            int i, j;

            if (fromCol > toCol)
                i = -1;
            else
                i = 1;


            if (fromLinea > toLinea)
                j = -1;
            else
                j = 1;



            fromCol += i;
            fromLinea += j;

            while (fromLinea != toLinea && fromCol != toCol)
            {
                int pos = colLineToPos(fromCol, fromLinea);
                int piece = tab[pos];

                if ((piece & ~REINA) == BLANCA)
                    fichasBlancas--;
                else if ((piece & ~REINA) == NEGRA)
                    fichasNegras--;

                tab[pos] = VACIA;
                fromCol += i;
                fromLinea += j;
            }
        }


        /// <sumary>
        ///   Inicializa el tablero para comenzar una nueva partida
        /// </sumary>
        public void inicializarTablero()
        {
            int i;


            fichasBlancas = 12;
            fichasNegras = 12;

            // Por defecto, primer turno para las fichas blancas
            jugadorConTurno = BLANCA;

            for (i = 0; i < 12; i++)
                tab[i] = NEGRA;

            for (i = 12; i < 20; i++)
                tab[i] = VACIA;

            for (i = 20; i < 32; i++)
                tab[i] = BLANCA;
        }

        /// <sumary>
        ///   Comprueba que el parmetro sea par
        /// </sumary>
        /// <param name="x">
        ///   Valor numrico
        /// </param>
        /// <value>
        ///   Devuelve true si el valor es par
        /// </value>
        private bool esPar(int x)
        {
            return x % 2 == 0;
        }


        /// <sumary>
        ///   Transforma una posicin columna/lnea en una posicin de tablero[32]
        /// </sumary>
        /// <param name="col">
        ///   Columna del tablero (0-7)
        /// </param>
        /// <param name="linea">
        ///   Lnea del tablero (0-7)
        /// </param>
        /// <value>
        ///   Posicin en el tablero
        /// </value>
        private int colLineToPos(int col, int linea)
        {
            if (esPar(linea))
                return linea * 4 + (col - 1) / 2;
            else
                return linea * 4 + col / 2;
        }


        /// <sumary>
        ///   Convierte posicin de tablero[32] en una fila
        /// </sumary>
        /// <param name="x">
        ///   Posicin en el tablero
        /// </param>
        /// <value>
        ///   Fila del tablero
        /// </value>
        private int posToFila(int x)
        {
            return x / 4;
        }


        /// <sumary>
        ///   Convierte posicin de tablero[32] en una columna
        /// </sumary>
        /// <param name="x">
        ///   Posicin en el tablero
        /// </param>
        /// <value>
        ///   Columna del tablero
        /// </value>
        private int posToCol(int x)
        {
            return (x % 4) * 2 + ((x / 4) % 2 == 0 ? 1 : 0);
        }

        /// <sumary>
        ///   Devuelve una cadena para mostrar en el turno
        /// </sumary>
        /// <value>
        ///   Color del jugador que tiene el turno
        /// </value>
        public String getColorTurno()
        {
            if (jugadorConTurno == BLANCA)
                return "blancas";
            else
                return "rojas";
        }
    }
}












